/* bfbtester -- Brute Force Binary Tester
 *
 * Copyright 2000	Mike Heffner <spock@techfour.net>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>


#include "datatypes.h"
#include "tempfiles.h"
#include "utils.h"


typedef struct _file_entry FileEntry;
struct _file_entry {

	char * filename;	/* excluding leading directories */

	int		statd;
	u_int32_t	fileno;
	struct stat	sb;

	struct _file_entry *next;

};


typedef struct _tempfile_watch_request TempFileRequest;
struct _tempfile_watch_request {

	pthread_mutex_t	mutex;
	pthread_cond_t	cond;
	int	count;	/* Number of requests - normally 0 or 1 */
	int status; /* 0 - none, 1 - start up, 2 - running, 3 - shutting down */

	TempfileCB	create_cb;	/* creation call back */
	TempfileCB	remove_cb;	/* remove call back */
	void		*cb_data;	/* user data */
};

static int initialized = FALSE;	/* Set to true if we are using tempfile
								 * monitoring
								 */
static FileEntry *head = NULL;
static TempFileRequest request = { PTHREAD_MUTEX_INITIALIZER, 
								   PTHREAD_COND_INITIALIZER,
								   0, 0, NULL, NULL, NULL };

static int
isNew(struct dirent *entry)
{
	FileEntry *entries = head;

	while(entries && entries->fileno != entry->d_fileno)
		entries = entries->next;

	if (entries)
		return FALSE;
	else
		return TRUE;
}

static void
addResult(struct dirent *entry)
{
	FileEntry *entries = head, *new;
	if (entries)
		while(entries->next) entries = entries->next;
	new = xmalloc(sizeof(FileEntry));
	
	new->filename = strdup(entry->d_name);
	new->fileno = entry->d_fileno;
	new->next = NULL;

	if (head)
		entries->next = new;
	else
		head = new;
	if (request.count)
		(*request.create_cb)(new->filename, &new->sb, request.cb_data);
}

static void
addIfNew(struct dirent *entry, int flags)
{
	if (isNew(entry))
		addResult(entry);
}

static void
tempfile_scandir(DIR *dir, int flags)
{
	struct dirent	entry;
	struct dirent	*result;
	int ret;

	ret = readdir_r(dir, &entry, &result);
	while (ret == 0 && result != NULL) {
		addIfNew(result, flags);
		ret = readdir_r(dir, &entry, &result);
	}
/*
	if (ret != 0)
		bfbt_message(MSG_TYPE_ERRNO_WARN, "readdir_r failed");
*/
	rewinddir(dir); /* rewinds for next time */
}

static void
tempfile_request_startup(void)
{
	/* Already have mutex locked */
	request.status = 2;
	pthread_cond_signal(&request.cond);
}

static void
tempfile_request_shutdown(void)
{
	/* Already have mutex locked */
	request.status = 0;
	pthread_cond_signal(&request.cond);
}

static void *
tempfile_watch(void *data)
{

	char *dirname = "/tmp";
	DIR	*dir;
	int	lasttime = 0;

	bfbt_message(MSG_TYPE_MESSAGE, "Starting tempfile monitor");
	initialized = TRUE;
	if ((dir = opendir(dirname)) == NULL)
		bfbt_message(MSG_TYPE_ERROR_ABORT, "opendir: %s failed", dirname);
	while(1){
		
		tempfile_scandir(dir, 0);

		pthread_mutex_lock(&request.mutex);
		if (request.status == 1 || request.status == 3) {
			if (request.status == 1) {	/* starting a new request */
				if (lasttime) {
					tempfile_request_startup();
					lasttime = 0;
				} else
					lasttime = 1;
			} else /* request.status == 3 */ {	/* shutting down request */
				if (lasttime) {
					tempfile_request_shutdown();
					lasttime = 0;
				} else
					lasttime = 1;
			}
		} else
			while (request.status == 0)
				pthread_cond_wait(&request.cond, &request.mutex);
		pthread_mutex_unlock(&request.mutex);
		pthread_testcancel();
	}
	return NULL;
}

int
tempfile_request_add(TempfileCB	create, TempfileCB remove, void *cbdata)
{
	if (!initialized)
		return 0;
	pthread_mutex_lock(&request.mutex);
	if (request.count || request.status != 0)
		return -1;
	request.count = 1;
	request.status = 1;
	pthread_cond_signal(&request.cond);
	while (request.status == 1)
		pthread_cond_wait(&request.cond, &request.mutex);
	request.create_cb = create;
	request.remove_cb = remove;
	request.cb_data = cbdata;
	pthread_mutex_unlock(&request.mutex);
	return 0;
}

int
tempfile_request_subtract(void)
{
	if (!initialized)
		return 0;
	pthread_mutex_unlock(&request.mutex);
	if (!request.count || request.status != 2)
		return -1;
	request.status = 3;
	while (request.status == 3)
		pthread_cond_wait(&request.cond, &request.mutex);
	request.count--;
	request.create_cb = NULL;
	request.remove_cb = NULL;
	request.cb_data = NULL;
	pthread_mutex_unlock(&request.mutex);
	return 0;
}

void
tempfile_start(THREAD_DATA *data)
{
	if (pthread_create(&data->tempfile_thread, NULL, tempfile_watch, NULL) != 0)
		bfbt_message(MSG_TYPE_ERRNO_ABORT, "pthread_create");
}
